/*
**	Apple Macintosh Developer Technical Support
**
**	Routines demonstrating how to play a sound with many of the same features as QuickTime.
**
**	by Mark Cookson, Apple Developer Technical Support
**
**	File:	SimpleApp Sound.c
**
**	Copyright 1996 Apple Computer, Inc.
**	All rights reserved.
**
**	You may incorporate this sample code into your applications without
**	restriction, though the sample code has been provided "AS IS" and the
**	responsibility for its operation is 100% yours.  However, what you are
**	not permitted to do is to redistribute the source as "Apple Sample
**	Code" after having made changes. If you're going to re-distribute the
**	source, we require that you make it clear in the source that the code
**	was descended from Apple Sample Code, but that you've made changes.
*/

#include "SimpleApp_Sound.h"

/*-----------------------------------------------------------------------*/
pascal	void	MyScrollAction (ControlHandle control, short part)
/*-----------------------------------------------------------------------*/
{
	long	position = GetControlValue (control);

	if (part) {
		switch (part) {
		case kInUpButtonControlPart:
			position -= oneBuffer;
			if (CDStyle == true) {
				if (gPlayBackwards == true) {
					ASoundPlayBackwards (mySoundInfo, true);
				}
			}
			break;
		case kInDownButtonControlPart:
			position += oneBuffer;
			break;
		case kInPageUpControlPart:
			position -= tenBuffers;
			if (CDStyle == true) {
				if (gPlayBackwards == true) {
					ASoundPlayBackwards (mySoundInfo, true);
				}
			}
			break;
		case kInPageDownControlPart:
			position += tenBuffers;
			break;
		default:
			DebugStr ("\pWhere did you click?");
		}
		SetControlValue (control, position);
		ASoundSetCurBuffer (mySoundInfo, position);

		if (CDStyle == true) {
			ASoundSetBytesCopied (mySoundInfo, position * ASoundGetBufferSize (mySoundInfo));
		}
	}
}

/*-----------------------------------------------------------------------*/
pascal	void	DoScroll (ControlHandle control, short part)
/*-----------------------------------------------------------------------*/
{
	long	position	= nil;
	Point	pt			= {nil, nil};
	OSErr	theError	= noErr;
	
	if (part != nil) {
		switch (part) {
			case kInIndicatorControlPart:
				if (CDStyle == false) {
					theError = ASoundPauseForAdjust (mySoundInfo);
				}
				GetMouse(&pt);
				part = TrackControl (control, pt, nil);
				position = GetControlValue (control);
				ASoundSetCurBuffer (mySoundInfo, position);
				ASoundSetBytesCopied (mySoundInfo, position * ASoundGetBufferSize (mySoundInfo));
				break;
			case kInUpButtonControlPart:
			case kInDownButtonControlPart:
			case kInPageUpControlPart:
			case kInPageDownControlPart: {
				ControlActionUPP actionProc = NewControlActionProc (MyScrollAction);
				if (CDStyle == false) {
					theError = ASoundPauseForAdjust (mySoundInfo);
				}
				part = TrackControl (control, pt, actionProc);
				DisposeRoutineDescriptor (actionProc);
				break;
			}
		}
	
		if (CDStyle == false) {
			position = GetControlValue (control);
			ASoundSetBytesCopied (mySoundInfo, position * ASoundGetBufferSize (mySoundInfo));
			theError = ASoundPauseForAdjust (mySoundInfo);
		}
	}
}

/*-----------------------------------------------------------------------*/
		void	MyIdleProc (EventRecord *evt)
/*-----------------------------------------------------------------------*/
{
#ifndef __SC__
#pragma unused (evt)
#endif
	long		curBuf		= nil;

	if (soundPlaying == true) {
		curBuf = ASoundGetCurBuffer (mySoundInfo);
		if (curBuf != nil) {
			SetControlValue (scrollBar, curBuf);
		}
	}

	if (ASoundIsDone (mySoundInfo) == true) {
		EnableControl (startButton);
		DisableControl (stopButton);
		DisableControl (resumeButton);
		DisableControl (scrollBar);
		SetControlValue (scrollBar, kScrollMinValue);
		ASoundDonePlaying (mySoundInfo, kCloseFile + kFreeMem);
		InstallIdleProc ((EventProcs)nil);
	}
}

/*-----------------------------------------------------------------------*/
		OSErr	DoPlay (void)
/*-----------------------------------------------------------------------*/
{
	OSErr	theErr;
	
	theErr = ASoundGetFileToPlay (mySoundInfo);
	if (theErr == noErr) {
		theErr = ASoundStartPlaying (mySoundInfo, nil);
		if (theErr == noErr) {
			SADisableMe();
			EnableControl(resumeButton);
			EnableControl(stopButton);
			paused = false;
			InstallIdleProc ((EventProcs)MyIdleProc);
			EnableControl (scrollBar);
			SetControlMaximum (scrollBar, ASoundGetNumBuffers (mySoundInfo) - 1);
		}
	}

	return theErr;
}

/*-----------------------------------------------------------------------*/
		OSErr	DoPause (void)
/*-----------------------------------------------------------------------*/
{
	Boolean			theError;
	
	theError = ASoundPause (mySoundInfo);

	if (paused == false) {
		paused = true;
		SetMyTitle ("\pResume Sound");
	}
	else {
		paused = false;
		SetMyTitle ("\pPause Sound");
	}

	return theError;
}

/*-----------------------------------------------------------------------*/
		OSErr	DoStop (void)
/*-----------------------------------------------------------------------*/
{
	OSErr theError;

	theError = ASoundStop (mySoundInfo);

	SADisableMe();
	EnableControl(startButton);
	DisableControl(resumeButton); 

	if (paused == true) {
		SetControlTitle(resumeButton,"\pPause Sound");
	}

	return theError;
}

/*-----------------------------------------------------------------------*/
pascal	short	DoPostGroupHit (long  modifiers,
								ControlRef theControl,
								ButtonItemRecHandle brh,
								short item)
/*-----------------------------------------------------------------------*/
{
#pragma unused (modifiers)
#pragma unused (theControl)
#pragma unused (brh)

	switch (item) {
		case qtStyle:
			DisableControl (checkBox);
			SetControlValue (checkBox, false);
			CDStyle = false;
			break;
		case cdStyle:
			EnableControl (checkBox);
			SetControlValue (checkBox, gPlayBackwards);
			CDStyle = true;
			break;
		default:
			DebugStr ("\pDoPostGroupHit got an invalid control passed to it!");
	}

	return nil;
}

/*-----------------------------------------------------------------------*/
		short	DoPlayBackwards (long modifiers)
/*-----------------------------------------------------------------------*/
{
#pragma unused (modifiers)

	gPlayBackwards = !gPlayBackwards;
	SetControlValue (checkBox, gPlayBackwards);

	return nil;
}

/*-----------------------------------------------------------------------*/
		void	main (void)
/*-----------------------------------------------------------------------*/
{
	Rect	r;
	short	err;
	Str255 names[] = {"\pQuickTime style",
					  "\pCD style"};

	InitSimpleApp (2, kUseStandardMenu);      	/* Simple App Sets up the Tool Box For us */
	GetDocumentWindow (128); 					/* Get our stored window */

	SetRectLocation (&r, 10, 2);				/* This Sets a rects anchor point */
	SetRectDimensions (&r, 90, 20);				/* This Sets its size without changing it position */
	{
		long buttonID = 1;
		err = InstallPushButton (&buttonID, gSACurrentWindow, "\pPlay Sound", &r, (char)0, (ButtonHitProc) DoPlay, (ButtonUpdateProc)nil);
		startButton = gSACurControl;
	}
	
	SetRectLocation (&r, 110, 2);				/* This Sets a rects anchor point */
	SetRectDimensions (&r, 100, 20);			/* This Sets its size without changing it position */
	{
		long buttonID = 2; 
		err = InstallPushButton (&buttonID, gSACurrentWindow, "\pPause Sound", &r, (char)0, (ButtonHitProc) DoPause, (ButtonUpdateProc)nil);
		resumeButton  = gSACurControl;
		DisableControl (resumeButton);
	}
	
	SetRectLocation (&r, 220, 2);				/* This Sets a rects anchor point */
	SetRectDimensions (&r, 90, 20);				/* This Sets its size without changing it position */
	{
		long buttonID = 3;
		err = InstallPushButton (&buttonID, gSACurrentWindow, "\pStop Sound", &r, (char)0, (ButtonHitProc) DoStop, (ButtonUpdateProc)nil);
		stopButton = (ControlHandle)SAGetObjectHandle(buttonID);	/* lets remember the control handle */
		DisableControl (stopButton);					/* disable the control */
	}

	SetRectDimensions (&r, 300, 16);
	SetRectLocation (&r, 10, 30);
	{
	 long scrollBarID  = 4;	
		err = InstallScrollBar (&scrollBarID, gSACurrentWindow, "\p", &r, (char)0, (MyActionProc)DoScroll, (MyActionProc)nil, (MyActionProc)nil);
		scrollBar = (ControlHandle)SAGetObjectHandle(scrollBarID);		/* get the object control handle */
		DisableControl (scrollBar);
		SetControlMinimum (scrollBar, kScrollMinValue);
		SetControlMaximum (scrollBar, kScrollMinValue);
		SetControlValue (scrollBar, kScrollMinValue);
	}

	SetRectLocation (&r, 20, 100);
	{
		long checkBoxID = 5;
		err = InstallCheckBox (&checkBoxID, gSACurrentWindow, "\pBackwards playing", &r, nil, false, (ButtonHitProc)DoPlayBackwards, (ButtonUpdateProc)nil);
		checkBox = (ControlHandle)SAGetObjectHandle(checkBoxID);
		DisableControl (checkBox);
	}

	SetRectLocation (&r, 10, 60);
	SetRectDimensions (&r, 130, 40);
	err	= InstallRadioGroup (6,					/* Group ID */
							gSACurrentWindow, 	/* owner window */
							2,					/* Number of items */
							"\pRadio Group",	/* Group Name */
							names, 				/* Array of button Names */
							&r, 				/* Group bounds */
							nil,				/* default item (zero based) */
							2, 					/* Horz spacing */
							2, 					/* vert spacing */
							16,					/* Button Height */
							130,				/* Button widths */
							nil,				/* Pre Hit Proc */
							DoPostGroupHit, 	/* Post Hit Proc */
							nil					/* Group Update Proc */
							);

	GetPrintArea (gSACurrentWindow,&r);			/* Set the print area top coordinate */
	r.top = 22;  								/* so that you don't scroll the button out of view */
	SetPrintArea (gSACurrentWindow,&r);
	mySoundInfo = ASoundNew (&err);

	if (err == noErr)
		Run (); 								/* Let SimpleApp handle the rest */
}

// utility functions since I already keep the control handle around
/*-----------------------------------------------------------------------*/
	void EnableControl (ControlRef cr)
/*-----------------------------------------------------------------------*/
{
	if (cr != nil) {
		HiliteControl(cr, nil);
	}
}

/*-----------------------------------------------------------------------*/
	void DisableControl (ControlRef cr)
/*-----------------------------------------------------------------------*/
{
	if (cr != nil) {
		HiliteControl(cr, 255);
	}
}
